#!/usr/bin/python3
"""
Koi is a Python-based CMS.

Koi uses a flat-file database to organize content stored in "koi"
files (which contain the page data in JSON format). It was developed
using Python 3.8.2+ and support for earlier versions is unknown.

To install, download from https://reimeika.ca and uncompress koi.zip
(this creates a new directory koi/). From a terminal, change into the
new directory and run ./koi.py. You can then go to:

http://localhost:8080

to start using the CMS.

Marco De la Cruz-Heredia (marco@reimeika.ca)
License: 3-clause BSD (see below)

Copyright (c) 2023, Marco De la Cruz-Heredia.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the author nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

__author__ = 'Marco De la Cruz-Heredia'
__version__ = '0.40'
__license__ = '3-clause BSD'

#############################################################################

import sys
import os
import logging
import glob
import json
import time
import tempfile
import secrets
import re
import hashlib
import string
import smtplib
import ssl
from email.message import EmailMessage

try:
    import bottle
    import config
except Exception as e:
    print(f'CRITICAL: koi: : cannot import required module [{e}]')
    raise SystemExit

# std_out goes to error_log, see:
# http://code.google.com/p/modwsgi/wiki/ApplicationIssues
logging.basicConfig(format=config.logformat,
                    level=getattr(logging, config.loglevel))

# https://stackoverflow.com/questions/61517/python-dictionary-from-an-objects-fields
CONFIG = dict((key, value) for key, value in config.__dict__.items() \
              if not key.startswith('__'))
CONFIG['koi_version'] = __version__

# LDAP was a later addition to koi.
if not CONFIG.get('ldap', False):
    CONFIG['ldap'] = {'auth': False, 'uri': None, \
                      'binddn': None, 'timeout': None}
elif CONFIG['ldap']['auth']:
    import ldap

pyver = sys.version_info
if pyver < (3, 8, 2):
    logging.error('unsupported python version, 3.8.2+ required')

if len(config.session_cookie_sig) < 32:
    logging.critical('invalid session_cookie_sig')
    raise SystemExit

try:
    # https://passlib.readthedocs.io/en/stable/narr/totp-tutorial.html
    from passlib.hash import sha512_crypt
    from passlib.totp import TOTP
    # "yescrypt" is new and not yet supported by passlib :(
    import crypt
    from hmac import compare_digest as compare_hash
    auth = True
except Exception as e:
    msg = f'could not import encryption modules, \
sessions/authentication disabled [{e}]'
    logging.error(msg)
    auth = False

koipy_file_rp = os.path.realpath(os.path.dirname(__file__))

logging.info(f'python version {pyver}')
logging.info(f'bottle version {bottle.__version__}')
logging.info(f'koi version {__version__}')
logging.debug(f'config is "{CONFIG}"')
logging.debug(f'koi.py path is "{koipy_file_rp}"')
if auth:
    logging.debug('sessions/authentication enabled')

template_dir_fp = os.path.join(koipy_file_rp, config.dir_templates)
logging.debug(f'template path is "{template_dir_fp}"')
bottle.TEMPLATE_PATH.append(template_dir_fp)

if config.dir_tmp:
    logging.debug(f'setting temp dir to "{config.dir_tmp}"')
    tempfile.tempdir = config.dir_tmp

koi_login = os.path.splitext(config.tpl_login)[0]+'.koi'
logging.debug(f'koi login file is "{koi_login}"')

#############################################################################

# https://en.wikipedia.org/wiki/List_of_HTTP_status_codes
@bottle.error(400)
def error400(error):
    """Bad Request"""
    logging.warning(f'[400 err] {error.body}')
    return bottle.template(config.tpl_error, \
                           CODE=400, \
                           DETAILS=error.body, \
                           CONFIG=CONFIG)

@bottle.error(403)
def error403(error):
    """Forbidden"""
    logging.debug(f'[403 err] {error.body}')
    return bottle.template(config.tpl_error, \
                           CODE=403, \
                           DETAILS=error.body, \
                           CONFIG=CONFIG)

@bottle.error(404)
def error404(error):
    """Not Found"""
    logging.debug(f'[404 err] {error.body}')
    return bottle.template(config.tpl_error, \
                           CODE=404, \
                           DETAILS=error.body, \
                           CONFIG=CONFIG)

@bottle.error(413)
def error413(error):
    """Payload Too Large"""
    logging.warning(f'[413 err] {error.body}')
    return bottle.template(config.tpl_error, \
                           CODE=413, \
                           DETAILS=error.body, \
                           CONFIG=CONFIG)

@bottle.error(500)
def error500(error):
    """Internal Server Error"""
    logging.critical(f'[500 err] {error.body}')
    return bottle.template(config.tpl_error, \
                           CODE=500, \
                           DETAILS=error.body, \
                           CONFIG=CONFIG)

#############################################################################

def json_io(file_fp, action, fatal, data=''):
    """
json_io(file_fp[str], action[str], fatal[bool], data[obj]='') -> data[obj]
                                                              -> [HTTPResponse]

    Read or write data from/to a JSON file depending on whether "action" is
    "r" or "w". If "fatal" is True a 500 error will be raised on I/O errors,
    otherwise an empty string if a read. File data is returned on a read, an
    empty string if a write.

"""
    fn = f'json_io({file_fp=}, {action=}, {fatal=}, <data>)'
    logging.debug(f'executing "{fn}"')

    if action == "r":
        try:
            with open(file_fp, "r", encoding='utf-8') as fd:
                data = json.load(fd)
        except Exception as e:
            data = ''
            msg = f'unable to load JSON file "{file_fp}" [{e}]'
            logging.error(msg)
            if fatal:
                raise bottle.HTTPError(500, 'error opening file')
    elif action == "w":
        # We write the JSON file into a temp file:
        # https://security.openstack.org/guidelines/dg_using-temporary-files-securely.html
        # and then copy it to its final destination. Using "json.dump()" directly can
        # corrupt the file if the device runs out of space, hence the first dump and the
        # subsequent copy.
        tmpfile = ''
        try:
            logging.debug(f'will write JSON file "{file_fp}"')
            # "dir" assures the file is created on the same file system
            # where "file" resides.
            (fd, tmpfile) = tempfile.mkstemp(dir=os.path.dirname(file_fp))
            with os.fdopen(fd, 'w', encoding='utf8') as tmp:
                tmp.write(json.dumps(data, ensure_ascii=False))
            # The "replace" operation is atomic.
            logging.debug(f'moving "{tmpfile}" to "{file_fp}"')
            os.replace(tmpfile, file_fp)
            data = ''
        except Exception as e:
            data = ''
            msg = f'unable to write JSON file "{file_fp}" [{e}]'
            logging.error(msg)
            if fatal:
                raise bottle.HTTPError(500, 'error writing file')
        finally:
            if os.path.isfile(tmpfile):
                os.remove(tmpfile)
    else:
        raise ValueError

    logging.debug('exiting json_io()')
    return data

#############################################################################

def index_pages(pages_dir_fp, profile):
    """
index_pages(pages_dir_fp[str], profile[dict]) -> (index[dict], tree[dict])

    Create an index of the site. Structure is:

    index[page1] = {'key1': 'value1', 'key2': 'value2'... }
    index[page2] = {'keyA': 'valueA', 'keyB': 'valueB'... }
    :

    where e.g. "key1" is "title" ("value1" is the actual title),
    "key2" is "date" ("value2" is the publication date), etc.
    "pageN" is the N-th directory under "config.dir_pages" in which
    a .koi file resides.

    In addition, create a site tree as so:

    tree[page1] = {'path': <path to page>,
                   'template': <koi template>,
                   'uri': <page URI>,
                   'files': <list of files/dirs under 'path'>,
                   'serve': <True if current user can access>}

"""
    logging.debug(f'executing "index_pages({pages_dir_fp=})"')

    index = {}
    tree = {}
    # A koi_file_fp is e.g.:
    #
    #                page
    #                ,--,
    # /www/koi/pages/main/article.koi
    # `------------------'`-----'
    #        path         template
    #
    # uri = /pages/main
    for koi_file_fp in glob.glob(os.path.join(pages_dir_fp, '*/*.koi')):
        path = os.path.dirname(koi_file_fp)
        files = [os.path.basename(i) for i in \
                 glob.glob(os.path.join(path, '*'))]
        template = os.path.splitext(os.path.basename(koi_file_fp))[0]
        page = os.path.basename(path)
        uri = f'/{config.dir_pages}/{page}'
        koi_data = json_io(koi_file_fp, action='r', fatal=False)
        if koi_data:
            index[page] = koi_data
        else:
            index[page] = {}
        (serve, msg) = check_acl_access(os.path.basename(koi_file_fp), '', \
                                        index[page], profile)
        tree[page] = {'path': path, 'template': template, \
                      'uri': uri, 'files': files, 'serve': serve}

    logging.debug('exiting index_pages()')
    return (index, tree)

#############################################################################

def index_users():
    """
index_users() -> users[dict]

    Create an index of all user profiles.

"""
    logging.debug('executing "index_users()"')

    users = {}
    for account_file_fp in glob.glob(os.path.join(config.dir_accounts_fp, \
                                                  '*.json')):
        user_data = json_io(account_file_fp, action='r', fatal=False)
        # If the JSON file has no "koi_version" field it's not a koi user.
        if not user_data or not user_data.get('koi_version', ''):
            continue
        else:
            users[user_data['user']] = user_data

    logging.debug('exiting index_users()')
    return users

#############################################################################

def check_os_access(pages_dir_fp, page, file):
    """
check_os_access(pages_dir_fp[str], page[str], file[str])
                   -> (request_page[str], request_file[str], koi_file_fp[str])
                   -> [HTTPResponse]

    Check OS-level access to the page (e.g. read permissions) and
    other potential issues or configurations which forbid serving
    the request (raising the appropriate HTTP response). If the
    request is feasible return the full path to the requested
    page, the requested file (these two being the same if a page
    is requested), and the full path to the .koi file. Note that
    symlinks are only allowed within the page directory (and any
    sub-directories).

"""
    fn = f'check_os_access({pages_dir_fp=}, {page=}, {file=})'
    logging.debug(f'executing "{fn}"')

    request_page_rp = os.path.realpath(os.path.join(pages_dir_fp, page))
    request_file_rp = os.path.realpath(os.path.join(request_page_rp, file))

    # If file = '' then request_page_rp and request_file_rp are the same
    logging.debug(f'requested "{request_file_rp}"')

    if not request_file_rp.startswith(pages_dir_fp):
        msg = 'will not serve request, possible path traversal'
        raise bottle.HTTPError(400, msg)

    if config.force_ssl and \
       bottle.request.environ.get('wsgi.url_scheme') != 'https':
        msg = 'force_ssl set, will not serve unencrypted'
        raise bottle.HTTPError(403, msg)

    if page.startswith(config.noserve_prefix) or \
       file.startswith(config.noserve_prefix):
        msg = 'will not serve due to noserve_prefix'
        raise bottle.HTTPError(403, msg)

    if page.endswith(config.noserve_suffix) or \
       file.endswith(config.noserve_suffix):
        msg = 'will not serve due to noserve_suffix'
        raise bottle.HTTPError(403, msg)

    if page and not os.path.isdir(request_page_rp):
        msg = 'page not found'
        raise bottle.HTTPError(404, msg)

    if file and not os.path.isfile(request_file_rp):
        msg = 'file not found'
        raise bottle.HTTPError(404, msg)

    if not os.access(request_file_rp, os.R_OK):
        msg = 'missing, or no read access to, page or file'
        raise bottle.HTTPError(403, msg)

    koi_file_fp = glob.glob(os.path.join(request_page_rp, '*.koi'))

    if len(koi_file_fp) == 0:
        msg = 'no .koi file found'
        raise bottle.HTTPError(403, msg)
    elif len(koi_file_fp) > 1:
        msg = f'multiple .koi files found "{koi_file_fp}"'
        logging.warning(msg)
        raise bottle.HTTPError(403, 'multiple .koi files found')
    else:
        koi_file_fp = koi_file_fp[0]
        logging.debug(f'found "{koi_file_fp}"')

    if file == os.path.basename(koi_file_fp):
        msg = 'will not serve the .koi file'
        raise bottle.HTTPError(403, msg)

    if not os.access(koi_file_fp, os.R_OK):
        msg = '.koi file is unreadable'
        raise bottle.HTTPError(403, msg)

    logging.debug('exiting check_os_access()')
    return (request_page_rp, request_file_rp, koi_file_fp)

#############################################################################

def generate_token(uid):
    """
generate_token(uid[str]) -> token[str]

    Generate a token.

"""
    logging.debug(f'executing "generate_token({uid=})"')

    token = ':'.join([uid, \
                      bottle.request.environ.get('REMOTE_ADDR'), \
                      f'{int(time.time())}', \
                      secrets.token_hex()])
    logging.debug(f'generated token "{token}"')

    logging.debug('exiting generate_token()')
    return token

#############################################################################

def validate_token(token):
    """
validate_token(token[str]) -> profile[dict]
                          -> [HTTPResponse]

    Return the user's profile, otherwise raises an HTTP error if the
    token is invalid (profile dictionary is empty if no token is given
    or is stale).

"""
    logging.debug(f'executing "validate_token({token=})"')

    if token:
        # Since the cookie is signed we trust the integrity of the token.
        (uid, ip, ts, sig) = token.split(':')

        account_file_fp = os.path.join(config.dir_accounts_fp, uid+'.json')
        profile = json_io(account_file_fp, action='r', fatal=True)

        # When we get a stale token after the user has logged out (happens
        # if using multiple browsers simultaneously).
        if not profile['token']:
            msg = f'stale token "{token}" received (not logged in)'
            logging.warning(msg)
            profile = {}
            logging.debug('exiting validate_token()')
            return profile

        if not secrets.compare_digest(profile['token'], token):
            msg = f'invalid token "{token}" (expected "{profile["token"]}")'
            logging.warning(msg)
            raise bottle.HTTPError(403, 'invalid token')

        # This shouldn't normally happen since the browser should delete
        # expired cookies but a malicious actor could be trying to re-use
        # them (e.g. session cookie theft).
        age = time.time()-int(ts)
        if age > config.session_timeout:
            msg = f'token expired {int(age-config.session_timeout)}s ago'
            logging.warning(msg)
            raise bottle.HTTPError(403, 'invalid token')

        # Not allowing roaming helps with the theft of session cookies.
        req_ip = bottle.request.environ.get('REMOTE_ADDR')
        if req_ip != ip and not config.session_allow_roaming:
            logging.warning(f'invalid IP "{req_ip}" (expected "{ip}")')
            raise bottle.HTTPError(403, 'invalid token')
    else:
        profile = {}

    logging.debug('exiting validate_token()')
    return profile

#############################################################################

def check_acl_access(koi_file, file, koi_data, profile):
    """
check_acl_access(koi_file[str], file[str], koi_data[dict], profile[dict])
                                                     -> (serve[bool], whynot[str])

    Returns a tuple with a boolean indicating whether a page or file should be
    served given its ACL (or lack thereof), and if not a string explaining why
    the item (i.e. page or file) cannot be served.

    An ACL is an entry in the .koi file with the following structure:

      acl = {'item1': {
                       'users': UACL1,
                       'groups': GACL1,
                       'ips':   IACL1,
                       'time':  TACL1},
             'item2': {
                       'users': UACL2,
                       'groups': GACL2,
                       'ips': IACL2,
                       'time': TACL2},

             'item3': {...} ... }}

    "UACLn" is either a single wildcard ('*') or a list of users
    who are allowed access (or blocked, if the user name is prepended
    but a "!") and can include a '*' wildcard (which allows access
    to all logged-in users, subject to the other restrictions).
    An empty list means no users are allowed access unless they
    belong to an allowed group. Using a single '*' allows anyone,
    logged-in or not, to access the item, subject only to the IP and
    time ACLs.

    "GACLn" is a list of groups (some possibly blocked, if prepended
    with a "!"). If the list is empty no groups are explicitly blocked
    or allowed.

    "IACLn" is a list of IPs and/or subnets (some possibly blocked,
    if prepended with a "!"). If '*' is used instead of a list then
    all IPs are allowed.

    "TACLn" is an epoch (which can be "0" to allow access at any time).

    Items lacking an ACL are always served.

    For example:

      acl = {'item1': {
                       'users': ['*', '!u3'],
                       'groups': [],
                       'ips': '*',
                       'time':  0},
             'item2': {
                       'users': ['u1', 'u2', 'u3'],
                       'groups': ['gA', 'gB'],
                       'ips': ['ip1', 'ip2', '!ip3'],
                       'time':  1637440490},

             'item3': {...} ... }}

    Each item is a serve-able file within the page. If the file is
    the "<template>.koi" file (which never gets served) its ACL is
    effectively that of the entire page and its contents (which inherit
    these permissions unless they are explicity stated e.g. "item2").
    The ACL controls who can access the item based on a user list, an IP
    address list, and a timestamp (in unix time) before which the request
    will not be granted. "users" and "ips" allow for a wildcard ("0" in
    the case of the timestamp). Prepending a "!" to a user, group, or
    IP/subnet blocks the request (these supercede "allow" pemissions
    and override subsets i.e. a user can be denied but the rest of their
    group allowed, but once the group is denied explicitly allowing
    a user from the group will not grant access permissions - same for
    networks, IPs within an allowed subnet can be restricted, but once
    a subnet is blocked individual IPs within cannot be allowed access).

    If "groups" is not '[]' and the visitor (if logged in) belongs
    to one of the groups they gain the same access rights as if they
    were in the "users" list. Wildcards are not allowed in the "groups"
    list.

    Examples:

    Allow full access to anyone, from anywhere, at any time, to the
    page and the all files in the page directory (this is equivalent
    of having no ACL entry at all):

    acl = {'article.koi': {'users': '*', 'groups': [], 'ips': '*',
                           'time': 0}}

    Same as above but only logged-in users can access the page and
    files within:

    acl = {'article.koi': {'users': ['*'], 'groups': [], 'ips': '*',
                           'time': 0}}

    Note that ['*'] is not possible for "groups" or "ips". Also, the
    presence of '*' in the "users" list supercedes the presence of any
    other users i.e. ['*', 'joe', 'ana'] and ['*'] are equivalent.

    Suppose the page "/private" uses a template called "mine".
    Only give "nao" access to "/private" (and files therein) from
    192.168.2.1 at any time:

    acl = {'mine.koi': {'users': ['nao'], 'groups': [],
                        'ips': ['192.168.2.1'], 'time': 0}}

    As above, but only giving access to all those in group "av"
    except for user "aika":

    acl = {'mine.koi': {'users': ['!aika'], 'groups': ['av'],
                        'ips': ['192.168.2.1'], 'time': 0}}

    Give all users access to "/preview" from Tue, Jun 23, 2020 15:58:31DST,
    but only allow "ゆま" and "sora" to download "final.pdf" anytime from
    the "127.0" subnet:

    acl = {'preview.koi': {'users': '*', groups: [], 'ips': '*',
                           'time': 1592942311},
           'final.pdf': {'users': ['ゆま', 'sora'], groups: [],
                         'ips': ['127.0'], 'time': 0}}

    Other files under "/preview" inherit the "preview.koi" ACL i.e. full
    access starting from Jun 23, 2020 15:58:31DST in this case.

"""
    fn = f'check_acl_access({koi_file=}, {file=}, <koi_data>, {profile=})'
    logging.debug(f'executing "{fn}"')

    # We make an exemption for the login page for when "index_pages()"
    # is called.
    if koi_file == koi_login:
        logging.debug('login page exempted from ACL')
        logging.debug('exiting check_acl_access()')
        return (True, '')

    if not (acl := koi_data.get('acl', {})):
        logging.debug('no ACL found')
        if config.force_acl:
            msg = 'ACL required'
            logging.error(msg)
            logging.debug('exiting check_acl_access()')
            return (False, msg)
        else:
            logging.debug('exiting check_acl_access()')
            return (True, '')
    elif koi_file not in acl:
        msg = 'improperly formed ACL'
        logging.error(msg)
        logging.debug('exiting check_acl_access()')
        return (False, msg)
    else:
        logging.debug('processing ACL')

    if not (item := file):
        item = koi_file

    # Files not in the ACL inherit the ACL of the .koi in the page.
    if file and file not in acl:
        acl[file] = acl[koi_file]

    logging.debug(f'ACL restrictions for "{item}" are "{acl[item]}"')
    logging.debug(f'access requested by "{profile.get("user", "<NOLOGIN>")}"')
    # https://en.wikipedia.org/wiki/De_Morgan%27s_laws
    if not (acl[item]['time'] == 0 or time.time() > acl[item]['time']):
        msg = 'item has not been released'
        logging.debug('exiting check_acl_access()')
        return (False, msg)
    if acl[item]['ips'] != '*':
        ip = bottle.request.environ.get('REMOTE_ADDR')
        blocked_ips = [i[1:] for i in acl[item]['ips'] if i[0] == '!']
        if ip in blocked_ips or any(ip.startswith(i) for i in blocked_ips):
            msg = 'access denied from your IP/subnet'
            logging.debug('exiting check_acl_access()')
            return (False, msg)
        allowed_ips = [i for i in acl[item]['ips'] if i[0] != '!']
        if not any(ip.startswith(i) for i in allowed_ips):
            msg = 'cannot access item from your IP/subnet'
            logging.debug('exiting check_acl_access()')
            return (False, msg)
    if acl[item]['users'] == '*':
        logging.debug('exiting check_acl_access()')
        return (True, '')
    # The only way to access an item outside a session is if "user" ACL
    # is set to '*', already served above.
    if not profile:
        msg = 'must log in to access'
        logging.debug('exiting check_acl_access()')
        return (False, msg)
    profile_groups = [i.lower() for i in profile['groups']]
    # All checks below are processed since negating a serve is irreversible.
    blocked_users = [i[1:].lower() for i in acl[item]['users'] \
                     if i[0] == '!']
    blocked_groups = [i[1:].lower() for i in acl[item]['groups'] \
                      if i[0] == '!']
    if profile['user'].lower() in blocked_users:
        msg = 'access denied to the user'
        logging.debug('exiting check_acl_access()')
        return (False, msg)
    if set(profile_groups).intersection(set(blocked_groups)):
        msg = 'access denied to the group'
        logging.debug('exiting check_acl_access()')
        return (False, msg)
    if '*' in acl[item]['users']:
        logging.debug('exiting check_acl_access()')
        return (True, '')
    allowed_users = [i.lower() for i in acl[item]['users'] if i[0] != '!']
    allowed_groups = [i.lower() for i in acl[item]['groups'] if i[0] != '!']
    # Having group access is equivalent to explicitly allowing the user.
    if set(profile_groups).intersection(set(allowed_groups)):
        allowed_users = [profile['user'].lower()]
    if profile['user'].lower() in allowed_users:
        logging.debug('exiting check_acl_access()')
        return (True, '')
    else:
        msg = 'user is not allowed to access item'
        logging.debug('exiting check_acl_access()')
        return (False, msg)

#############################################################################

def get_upload(up_request):
    """
get_upload(up_request[bottle.FileUpload]) -> upload[dict]

    Process a bottle.FileUpload instance and retrieve the
    upload. If the upload is successful returns an "upload"
    dictionary containing the keys "OK" (boolean), "status",
    "content_type", "raw_filename", "safe_name", and
    "file_data" (whose value is a binary string containing
    the uploaded file). If the upload is not successful due
    to exceeding the upload size (config.upload_max_size)
    then "OK" is "False" and "file_data" is truncated.

    See also:
    http://bottlepy.org/docs/0.12/api.html#bottle.FileUpload

"""
    logging.debug('executing "get_upload(<bottle.FileUpload>)"')

    punctuation = re.compile(r'[^a-zA-Z0-9][^a-zA-Z0-9_-]*')
    upload = {}
    data_blocks = []
    byte_count = 0
    msg = 'receiving "{0}" file "{1}"'.\
        format(up_request.content_type, up_request.raw_filename)
    logging.debug(msg)
    upload['OK'] = True
    upload['status'] = 'upload successful'
    # We load it manually to monitor file size.
    while buf := up_request.file.read(config.upload_chunk_size):
        byte_count+=len(buf)
        if byte_count > config.upload_max_size:
            upload['OK'] = False
            msg = f'upload exceeds {config.upload_max_size} bytes'
            upload['status'] = msg
            break
        data_blocks.append(buf)
    upload['content_type'] = up_request.content_type
    upload['raw_filename'] = up_request.raw_filename
    raw = os.path.splitext(upload['raw_filename'])
    # Remove any non alphanumeric characters from the file name
    # (but keep the file extension, if any). Also limit the total
    # file name length and change a "koi" extension to "koi.up".
    safe_name = [punctuation.sub('', raw[0][:200]), \
                 punctuation.sub('', raw[1][1:10])]
    if not safe_name[0]:
        safe_name[0] = f'null-{int(time.time())}'
    if not safe_name[1]:
        safe_name[1] = 'null'
    elif safe_name[1] == 'koi':
        safe_name[1] += '.up'
    upload['safe_name'] = '.'.join(safe_name)
    logging.debug(f'"upload" ("file_data" not shown) is "{upload}"')
    upload['file_data'] = b''.join(data_blocks)
    logging.debug(f'uploaded {len(upload["file_data"])} bytes')

    logging.debug('exiting get_upload()')
    return upload

#############################################################################

def email_nonce(email):
    """
email_nonce(email[str]) -> nonce[int]

    Create a nonce and email it to the user for two-factor authentication.

"""
    logging.debug(f'executing "email_nonce({email=})"')

    nonce = ''.join(secrets.choice(string.digits) for i \
                    in range(config.twoF_email_digits))
    try:
        em = EmailMessage()
        em['Subject'] = config.twoF_email_subject
        em['From'] = config.twoF_email_from
        em['To'] = email
        em.set_content(config.twoF_email_msg.format(nonce=nonce))
        es = smtplib.SMTP(host=config.smtp_server, port=config.smtp_port)
        msg = 'sending 2-factor email to "{0}" via "{1}:{2}"'.\
            format(email, config.smtp_server, config.smtp_port)
        logging.debug(msg)
        if config.smtp_TLS:
            context = ssl.create_default_context()
            es.starttls(context=context)
        if config.smtp_login:
            logging.debug(f'SMTP auth as "{config.smtp_login}"')
            es.login(config.smtp_login, config.smtp_passwd)
        es.send_message(em)
        es.quit()
    except Exception as e:
        logging.debug(f'error sending 2-factor email [{e}]')
        raise bottle.HTTPError(500, 'cannot authenticate')

    logging.debug('exiting email_nonce()')
    return nonce

#############################################################################

def session(query):
    """
session(query[dict]) -> (err[str], profile[dict], cookie[str])
                     -> [HTTPResponse]

    Session handling function for login, logout, 2F and CSRF. Returns
    the profile (if possible), an error string (if any occur,
    otherwise the string is empty), and whether a cookie should be "set",
    "del", or "" (which does nothing).

"""
    logging.debug('executing "session(<query>)"')

    if not auth:
        raise bottle.HTTPError(400, 'cannot authenticate for sessions')

    if bottle.request.environ.get('wsgi.url_scheme') != 'https' and \
       config.loglevel != 'DEBUG':
        msg = 'non-encrypted login not allowed unless in DEBUG mode'
        raise bottle.HTTPError(403, msg)

    err = ''
    action = query.get('koi_session', '')

    if not action:
        # In this case we're just showing the initial login page, so nothing
        # to do.
        logging.debug('exiting session()')
        return (err, {}, '')
    elif action not in ['login', 'logout', 'chknonce']:
        logging.warning(f'unknown session action "{action}"')
        raise bottle.HTTPError(400, 'unknown action')

    logging.debug(f'processing session action "{action}"')

    if action == 'login':
        user = query.get('user', '')
        password = query.get('password', '')

        logging.debug(f'login attempt by user "{user}"')

        # All information submitted should be enforced by the form
        # client-side, so if anything is missing we end it here.
        if not user or not password:
            raise bottle.HTTPError(400, 'bad login request')

        # Internally all users are identified by their lowercased user name.
        uid = hashlib.md5(user.lower().encode()).hexdigest()
        account_file_fp = os.path.join(config.dir_accounts_fp, uid+'.json')
        logging.debug(f'seeking "{user}" profile at "{account_file_fp}"')
        if not os.path.isfile(account_file_fp):
            err = 'incorrect credentials'
            logging.debug(f'exiting session() with error "{err}"')
            return (err, {}, '')

        profile = json_io(account_file_fp, action='r', fatal=True)

        # Login attempt throttler. "knock_tss[-1]" is the timestamp of
        # the last attempted login. Login delay increases with the number
        # of attempts (if "config.throttle_attempts" is enabled, otherwise
        # len(profile['knock_tss']) is always 1).
        delay = config.throttle_delay*len(profile['knock_tss'])
        if (wait :=  profile['knock_tss'][-1] + delay - time.time()) > 0:
            err = f'login attempt rate exceeded, try again in {round(wait)}s'
            logging.debug(f'exiting session() with error "{err}"')
            return (err, {}, '')

        if config.throttle_attempts < 0:
            profile['knock_tss'] = [int(time.time())]
            json_io(account_file_fp, action='w', fatal=True, data=profile)
        elif len(profile['knock_tss']) > config.throttle_attempts:
            if config.throttle_lockout < 0:
                profile['locked'] = True
                profile['knock_tss'] = [profile['knock_tss'][-1]]
                json_io(account_file_fp, action='w', fatal=True, data=profile)
                err = 'account locked due to too many failed login attempts'
                logging.debug(f'exiting session() with error "{err}"')
                return (err, {}, '')
            elif (wait := profile['knock_tss'][-1] + config.throttle_lockout - \
                  time.time()) > 0:
                err = f'too many failed login attempts, try again in {round(wait)}s'
                logging.debug(f'exiting session() with error "{err}"')
                return (err, {}, '')
            else:
                # If the account has not been locked and enough time has
                # passed reset the counter.
                profile['knock_tss'] = [profile['knock_tss'][-1]]
                json_io(account_file_fp, action='w', fatal=True, data=profile)
        else:
            profile['knock_tss'].append(int(time.time()))
            json_io(account_file_fp, action='w', fatal=True, data=profile)

        if CONFIG['ldap']['auth']:
            ldap_bind = False
            try:
                l = ldap.initialize(CONFIG['ldap']['uri'])
                ldap_bind = True
                l.set_option(ldap.OPT_NETWORK_TIMEOUT, CONFIG['ldap']['timeout'])
                binddn = CONFIG['ldap']['binddn'].replace('%u', user)
                logging.debug(f'LDAP {binddn=}')
                l.simple_bind_s(binddn, password)
            except ldap.INVALID_CREDENTIALS:
                auth_pass = False
            except Exception as e:
                logging.error(f'LDAP error [{e}]')
                auth_pass = False
            else:
                auth_pass = True
            if ldap_bind:
                l.unbind()
        elif profile['hash'].startswith('$6$'):
            auth_pass = sha512_crypt.verify(password, profile['hash'])
        elif profile['hash'].startswith('$y$'):
            auth_pass = compare_hash(crypt.crypt(password, profile['hash']),
                                     profile['hash'])
        elif profile['hash'].startswith('!'):
            auth_pass = False
        else:
            # This should never happen...
            err = 'unsupported password hash'
            logging.debug(f'exiting session() with error "{err}"')
            return (err, {}, '')

        if not auth_pass:
            err = 'incorrect credentials'
            logging.debug(f'exiting session() with error "{err}"')
            return (err, {}, '')

        if profile['locked']:
            err = 'account is locked'
            logging.debug(f'exiting session() with error "{err}"')
            return (err, {}, '')

        if profile['multiF_check']['enabled']:
            # The TOTP is determined at login.
            n1 = n2 = 0
            if (twoF_email := profile['multiF_check']['twoF_email']):
                n1 = email_nonce(twoF_email)
            if profile['multiF_check']['twoF_totp']:
                n2 = 1
            profile['nonce'] = f'{n1}-{n2}'
            logging.debug(f"setting nonce: {profile['nonce']}")
        else:
            profile['nonce'] = ''
            profile['ip'] = bottle.request.environ.get('REMOTE_ADDR')
            profile['knock_tss'] = [profile['knock_tss'][-1]]
            profile['login_ts'] = int(time.time())

        profile['token'] = generate_token(uid)
        profile['xCSRF'] = secrets.token_hex()
        json_io(account_file_fp, action='w', fatal=True, data=profile)
        logging.debug('exiting session()')
        return(err, profile, 'set')

    token = bottle.request.get_cookie('token', \
                                      secret=config.session_cookie_sig)
    profile = validate_token(token)
    if profile:
        account_file_fp = os.path.join(config.dir_accounts_fp, \
                                       profile['uid']+'.json')

    if action == 'logout':
         # If no profile the user wasn't logged in to begin with
         # (or their token expired).
        if profile:
            profile['knock_tss'] = [profile['knock_tss'][-1]]
            profile['logout_ts'] = int(time.time())
            profile['token'] = ''
            profile['xCSRF'] = ''
            json_io(account_file_fp, action='w', fatal=True, data=profile)
        logging.debug('exiting session()')
        return (err, {}, 'del')

    if action == 'chknonce':
        # In case the cookie expired before 2F was completed.
        if not profile:
            logging.debug('could not retrieve profile (expired token)')
            raise bottle.HTTPError(400, 'cannot authenticate')
        if not profile['multiF_check']['enabled']:
            logging.debug('multi-factor not enabled, cannot validate nonce')
            raise bottle.HTTPError(400, 'cannot authenticate')
        if not profile['nonce']:
            logging.debug('cannot check non-existent nonce')
            raise bottle.HTTPError(400, 'cannot authenticate')
        n1 = n2 = ''
        if profile['multiF_check']['twoF_email']:
            n1 = profile['nonce'].split('-')[0]
        # The TOTP data is stored as a JSON string in "twoF_totp".
        if (twoF_totp := profile['multiF_check']['twoF_totp']):
            n2 = TOTP.using(secrets=config.twoF_totp_secret).from_source(twoF_totp).generate().token
        sys_nonce = int(f'{n1}{n2}')
        # An input nonce can be either one integer, or two separated by a dash
        # (which will become one long integer when that single dash is removed).
        OK = True
        try:
            usr_nonce = int(query.get('nonce').replace('-', '', 1))
        except:
            logging.debug('invalid nonce')
            OK = False
        logging.debug(f'comparing usr_nonce={usr_nonce} with sys_nonce={sys_nonce}')
        if OK and usr_nonce == sys_nonce and \
           secrets.compare_digest(profile['token'], token):
            msg = '2-factor nonce was successfully verified for "{0}"'.\
                format(profile['user'])
            logging.debug(msg)
            profile['nonce'] = ''
            profile['knock_tss'] = [profile['knock_tss'][-1]]
            profile['login_ts'] = int(time.time())
            profile['ip'] = bottle.request.environ.get('REMOTE_ADDR')
            json_io(account_file_fp, action='w', fatal=True, data=profile)
            cookie = 'set'
        else:
            msg = '2-factor nonce authentication failed for "{0}"'.\
                format(profile['user'])
            logging.debug(msg)
            profile['nonce'] = ''
            profile['token'] = ''
            profile['xCSRF'] = ''
            json_io(account_file_fp, action='w', fatal=True, data=profile)
            err = '2-factor authentication failed'
            profile = {}
            # A delete_cookie cookie response does not work if it's
            # followed by raising an HTTPError.
            cookie = 'del'
        logging.debug('exiting session()')
        return(err, profile, cookie)

#############################################################################

@bottle.route('/', method=['GET', 'POST'])
@bottle.route(f'/{config.dir_pages}/<page:re:.*/?>', method=['GET', 'POST'])
@bottle.route(f'/{config.dir_pages}/<page>/<file>')
def process_page(page=config.page_home, file=''):
    """
process_page(page[str], file[str]) -> [HTTPResponse]

    Serve a web request for a page, file, or an upload.

"""
    logging.debug(f'executing "process_page({page=}, {file=})"')
    logging.debug(f'request method is "{bottle.request.method}"')

    pages_dir_fp = os.path.join(koipy_file_rp, config.dir_pages)
    (request_page_rp, request_file_rp, koi_file_fp) = \
        check_os_access(pages_dir_fp, page, file)

    template = os.path.splitext(os.path.basename(koi_file_fp))[0]
    me = {'page': page, 'path': request_page_rp, 'template': template, \
          'koi_fp': koi_file_fp, 'uri': f'/{config.dir_pages}/{page}', \
          'files': [os.path.basename(i) for i in \
                    glob.glob(os.path.join(request_page_rp, '*'))]}

    koi_data = json_io(koi_file_fp, action='r', fatal=True)

    # http://bottlepy.org/docs/0.12/tutorial.html#request-data
    # http://bottlepy.org/docs/0.12/api.html#bottle.BaseRequest.params
    # http://bottlepy.org/docs/0.12/api.html#bottle.FormsDict
    # http://bottlepy.org/docs/0.12/tutorial.html#html-form-handling
    try:
        query = dict(bottle.request.params.decode())
    except Exception as e:
        query = {}
        logging.warning(f'cannot parse query input [{e}]')

    if page == config.page_login:
        logging.debug('session transaction, query dictionary not shown')
        (err, profile, cookie) = session(query)
        if err:
            logging.debug(f'error "{err}" negotiating session')
        if cookie == 'del':
            bottle.response.delete_cookie('token',  \
                                          path='/', \
                                          secret=config.session_cookie_sig)
        elif cookie == 'set':
            # When the cookie reaches "config.session_timeout" it should
            # simply cease to exist.
            bottle.response.set_cookie(name='token', \
                                       path='/', \
                                       value=profile.get('token', ''), \
                                       secure=config.force_ssl, \
                                       httponly=True, \
                                       max_age=config.session_timeout, \
                                       secret=config.session_cookie_sig)
        logging.debug('exiting process_page()')
        return bottle.template(config.tpl_login,  \
                               BOTTLE=bottle.request.environ, \
                               CONFIG=CONFIG,   \
                               ERR=err,         \
                               ME=me,           \
                               PAGE=koi_data,   \
                               PROFILE=profile, \
                               QUERY=query)
    else:
        logging.debug(f'GET/POST query dictionary is "{query}"')
        token = bottle.request.get_cookie('token', \
                                          secret=config.session_cookie_sig)
        profile = validate_token(token)
        # We don't check "page_assets" against the nonce since it normally
        # contains css/images/etc required by the login page (if the browser
        # tries to access a file in that page before the nonce is submitted
        # it'd get wiped out, a common occurence due to aggressive pre-fetching).
        if profile and profile['nonce'] and page != config.page_assets:
            logging.warning(f'attempt to bypass 2-factor by "{profile["user"]}"')
            profile['nonce'] = ''
            profile['token'] = ''
            profile['xCSRF'] = ''
            account_file_fp = os.path.join(config.dir_accounts_fp, \
                                           profile['uid']+'.json')
            json_io(account_file_fp, action='w', fatal=True, data=profile)
            profile = {}
            bottle.response.delete_cookie('token',  \
                                          path='/', \
                                          secret=config.session_cookie_sig)
            # Uncommenting the following will not delete the cookie, so we
            # follow through to "check_acl_access" and depending of the ACL
            # perhaps kick the user back to the login page.
            # raise bottle.HTTPError(403, 'security violation detected')

        # Although "page_assets" is not nonce-checked its ACL is still
        # enforced (though not making it fully accessible will likely cause
        # issues unless the login process makes no use of its contents).
        (serve, msg) = check_acl_access(os.path.basename(koi_file_fp), file, \
                                        koi_data, profile)
        if not serve:
            if msg == 'must log in to access':
                if file:
                    wants = f'{page}/{file}'
                else:
                    wants = page
                query['koi_wants'] = wants
                query['koi_stay'] = ''
                koi_login_fp = os.path.join(pages_dir_fp, config.page_login, \
                                            koi_login)
                koi_data = json_io(koi_login_fp, action='r', fatal=True)
                logging.debug('exiting process_page()')
                return bottle.template(config.tpl_login,  \
                                       BOTTLE=bottle.request.environ, \
                                       CONFIG=CONFIG,   \
                                       ERR='',          \
                                       ME=me,           \
                                       PAGE=koi_data,   \
                                       PROFILE={},      \
                                       QUERY=query)
            else:
                raise bottle.HTTPError(403, msg)

        upload_requests = bottle.request.files.getall('koi_file_upload')
        if upload_requests:
            num_up = len(upload_requests)
            logging.debug(f'<bottle.FileUpload> detected (x{num_up})')

    if profile and (query or upload_requests):
        xCSRF = query.get('xCSRF', '')
        if xCSRF:
            logging.debug(f'xCSRF token found "{xCSRF}"')
            if not secrets.compare_digest(xCSRF, profile.get('xCSRF', '')):
                logging.warning(f'attempted CSRF by "{profile["user"]}"')
                raise bottle.HTTPError(403, 'security violation detected')
        else:
            msg = 'will not serve in-session request without xCSRF'
            raise bottle.HTTPError(403, msg)

    if file:
        logging.debug(f'serving static file "{file}" at "{request_file_rp}"')
        logging.debug('exiting process_page()')
        return bottle.static_file(file, root=request_page_rp)

    template_file = os.path.join(template_dir_fp, template) + '.tpl'
    if not os.access(template_file, os.R_OK):
        msg = f'missing or no read access to template file "{template_file}"'
        logging.error(msg)
        raise bottle.HTTPError(404, 'template not found')
    else:
        logging.debug(f'using template file "{template_file}"')

    index = {}
    tree = {}
    if koi_data.get('get_index', False):
        (index, tree) = index_pages(pages_dir_fp, profile)
    else:
        logging.debug('page did not request an index/tree')

    users = {}
    if koi_data.get('get_users', False):
        users = index_users()
    else:
        logging.debug('page did not request users')

    upload = {}
    if upload_requests:
        if num_up > config.upload_max_files:
            msg = 'will not upload over {0} files at a time'.\
                format(config.upload_max_files)
            raise bottle.HTTPError(413, msg)
        n = 0
        for up_request in upload_requests:
            upload[n] = get_upload(up_request)
            n+=1

    logging.info(f'serving page "{request_page_rp}"')
    logging.debug('exiting process_page()')
    return bottle.template(template, \
                           BOTTLE=bottle.request.environ, \
                           CONFIG=CONFIG,   \
                           INDEX=index,     \
                           FORMS=bottle.request.forms.decode(), \
                           ME=me,           \
                           PAGE=koi_data,   \
                           PROFILE=profile, \
                           QUERY=query,     \
                           TREE=tree,       \
                           UPLOAD=upload,   \
                           USERS=users)

#############################################################################

if __name__ == '__main__':
    # Setting "reloader=True" will import modules twice... this does not
    # seem to be a problem, but in any case it's not a good idea to run
    # koi using the bottle server (it's really only for debugging).
    # The advantage of the reloader is that koi will automatically
    # restart every time this file, template or config is saved.
    bottle.run(host='localhost', port=config.local_http_port, \
               reloader=True, debug=True)
