#!/usr/bin/python3
"""
Koi account management. Run without arguments for an interactive
session, otherwise use "-h" to see command-line options.
"""

#############################################################################

def get_profiles():
    """
get_profiles() -> profiles[dict]

    Read all JSON profiles into a dictionary. Each dictionary
    key corresponds to a user's UID.
"""
    profiles = {}
    for i in glob.glob(os.path.join(config.dir_accounts_fp, '*.json')):
        with open(i, 'r', encoding='utf-8') as fd:
            data = json.load(fd)
        if not data.get('koi_version', ''): continue
        profiles[data['uid']] = data
    return profiles

#############################################################################

def get_password(user):
    """
get_password(user[str]) -> hash[str]
                        -> '!'

    Query for a password and return the corresponding hash.
"""
    while True:
        pr = f'\nPassword for {user} (press "enter" to disallow logins): '
        p1 = getpass.getpass(prompt=pr)
        if not p1:
            return '!'
        p2 = getpass.getpass(prompt='Confirm password: ')
        if p1 == p2:
            return sha512_crypt.hash(p1)
        else:
            print('Passwords do not match, please try again.')

#############################################################################

def get_email():
    """
get_email() -> email[str]

    Query for an email address.
"""
    while True:
        e1 = input("\nUser's email address: ")
        e2 = input('Confirm email: ')
        if e1 == e2:
            return e1
        else:
            print('Emails do not match, please try again.')

#############################################################################

def search_accounts():
    """
search_accounts()

    Search accounts for a string.

"""
    if not profiles:
        print('\nThere are currently no accounts.')
        return
    else:
        print(f'\nSearching {len(profiles)} account(s).')
    query = input('\nInput a string (case-insensitive): ')
    hits = []
    for uid, data in profiles.items():
        for i in data:
            if i in ['full_name', 'email', 'phone', 'address']:
                for key, val in data[i].items():
                    if query.lower() in val.lower():
                        hits.append(uid)
                        break
            elif i in ['name', 'user']:
                if query.lower() in data[i].lower():
                    hits.append(uid)
                    break
    if not len(hits):
        print ('\nNo matches found')
    elif len(hits) == 1:
        print ('\nFound one match below:\n')
        pprint.pprint(profiles[hits[0]])
    else:
        print ('\nFound the following matches:')
        for uid in hits:
            pf = profiles[uid]
            if pf['multiF_check']['enabled']:
                status = 'enabled'
            else:
                status = 'disabled'
            print(f'{pf["user"]}: name={pf["name"]}, locked={pf["locked"]}, \
multi-factor={status}, roles={pf["roles"]}, uid={uid}')
    return

#############################################################################

def add_account(interactive=True, args=None):
    """
add_account(interactive[bool], args[namespace])

    Add an account. Note that although case is preserved in user names
    it is ignored by "koi.py". Thus "aika" and "AIKA" user names are
    stored as such, but internally they are both "aika" (and hence they
    cannot co-exist). The logic behind this is for potential self-registration,
    particular if using email addresses, since users pay little attention
    to case. Note that groups are also internally lowercased (and this program
    will store them as such, so unlike user names they are not case-aware).

"""
    import re
    valid = re.compile(config.user_re)
    if interactive:
        print('\nAdding account')
        user = input('\nUser name (<enter> to cancel): ')
        nl = '\n'
    else:
        user = args.user
        nl = ''
    if not user:
        return
    uid = hashlib.md5(user.lower().encode()).hexdigest()
    if uid in profiles:
        print(f'{nl}That account already exists.')
        if interactive:
            return
        else:
            sys.exit(1)
    if not valid.match(user):
        print(f'{nl}That is not a valid user name, must match: {config.user_re}')
        if interactive:
            return
        else:
            sys.exit(1)
    if interactive:
        print (f'\nuid will be: {uid}')
    account_file_fp = os.path.join(config.dir_accounts_fp, uid+'.json')
    profile = {'user': user, 'uid': uid,  \
               'full_name': {'first': '', 'middle': '', 'last': ''}, \
               'gecos': '', 'token': '', 'nonce': '', \
               'xCSRF': '', 'data': {}, 'address': {'primary': ''}, \
               'email': {'primary': ''}, 'phone': {'primary': ''}, \
               'locked': False, 'roles': {config.tpl_edit: []}, \
               'cache': {}, 'login_ts': 0, 'logout_ts': 0, 'ip': None, \
               'knock_tss': [0], 'koi_version': __version__, \
               'created': int(time.time())}
    if not interactive:
        profile['roles'][config.tpl_edit] = ['editor'] if args.editor_on \
            else []
        profile['full_name']['first'] = '' if not args.firstname \
            else args.firstname
        profile['full_name']['middle'] = '' if not args.middlename \
            else args.middlename
        profile['full_name']['last'] = '' if not args.lastname \
            else args.lastname
        profile['address']['primary'] = '' if not args.primary_address \
            else args.primary_address
        profile['email']['primary'] = '' if not args.primary_email \
            else args.primary_email
        profile['phone']['primary'] = '' if not args.primary_phone \
            else args.primary_phone
        if args.name:
            profile['name'] = args.name
        else:
            profile['name'] = ' '.join(f'{profile["full_name"]["first"]} {profile["full_name"]["middle"]} {profile["full_name"]["last"]}'.split())
        profile['hash'] = '!' if not args.password \
            else args.password
        pg = [] if not args.group else [i.strip().lower() for \
                                         i in args.group]
        if args.agroup:
            pg += [i.strip().lower() for i in args.agroup]
        profile['groups'] = sorted(list(set(pg)))
        profile['multiF_check'] = {'twoF_email': '', 'twoF_totp': ''}
        if args.twof_email:
            profile['multiF_check']['enabled'] = True
            profile['multiF_check']['twoF_email'] = args.twof_email
        else:
            profile['multiF_check']['enabled'] = False
    else:
        profile['name'] = ''
        pwhash = ''
        if platform.system() == 'Linux':
            if user in [i[0] for i in pwd.getpwall()]:
                print(f'\nUser "{user}" found on the system.')
                reuse = 'n'
                if getpass.getuser() == 'root':
                    pr = "Import the GECOS field and shadow password [y/N]? "
                    reuse = input(pr)
                    pr = 'Extract "name" from the first GECOS field [y/N]? '
                    name_from_gecos = input(pr)
                else:
                    print('Note that you can use the shadow password via copy/paste')
                    print('or by running this program as root.\n')
                if reuse.lower() == 'y':
                    pwhash = spwd.getspnam(user).sp_pwdp
                    if not pwhash.lstrip('!').startswith('$6$') and \
                       not pwhash.lstrip('!').startswith('$y$'):
                        print(f'"{user}" account does not use either SHA-512 or \
yescrypt, locking.')
                        pwhash = '!'
                        locked = True
                    profile['gecos'] = pwd.getpwnam(user).pw_gecos
                    if name_from_gecos.lower() == 'y':
                        profile['name'] = profile['gecos'].split(',')[0]
        print('\nIf you specify any names below then the "name" field \
will be updated accordingly (unless imported from the GECOS field).')
        profile['full_name']['first'] = input('User first name: ')
        profile['full_name']['middle'] = input('User middle name(s): ')
        profile['full_name']['last'] = input('User last name: ')
        if not profile['name']:
            profile['name'] = ' '.join(f'{profile["full_name"]["first"]} {profile["full_name"]["middle"]} {profile["full_name"]["last"]}'.split())
        profile['address']['primary'] = input('User primary address: ')
        profile['email']['primary'] = input('User primary email: ')
        profile['phone']['primary'] = input('User primary phone: ')
        if not pwhash:
            pwhash = get_password(user)
        profile['hash'] = pwhash
        groups = input('Space-separated list of groups (<enter> for none): ')
        profile['groups'] = sorted(list(set([i.strip().lower() for \
                                             i in groups.split()])))
        pr = 'Register an email for 2-factor authenticaction [y/N]? '
        twoF_do = input(pr)
        if twoF_do.lower() == 'y':
            twoF_email = get_email()
            profile['multiF_check'] = {'enabled': True, 'twoF_totp': ''}
            profile['multiF_check']['twoF_email'] = twoF_email
        else:
            profile['multiF_check'] = {'enabled': False, 'twoF_email': '', \
                                       'twoF_totp': ''}
        print(f'\nSaving "{user}" in {account_file_fp}')
    with open(account_file_fp, 'w', encoding='utf-8') as fd:
        json.dump(profile, fd, ensure_ascii=False)
    os.chmod(account_file_fp, config.wwwfperms)
    if getpass.getuser() == 'root':
        shutil.chown(account_file_fp, config.wwwuid, config.wwwgid)
    if not interactive and not args.quiet:
        print(f'Account "{user}" has been added.')
        sys.exit()
    return

#############################################################################

def rem_account(interactive=True, user=None, action=None):
    """
rem_account(interactive[bool], user[str])

    Remove an account.

"""
    if interactive:
        print('\nRemoving account')
        user = input('\nUser name (<enter> to cancel): ')
        nl = '\n'
    else:
        user = args.user
        nl = ''
    if not user:
        return
    uid = hashlib.md5(user.lower().encode()).hexdigest()
    if uid not in profiles:
        print(f'{nl}The account "{user}" does not exist.')
        if interactive:
            return
        else:
            sys.exit(1)
    account_file_fp = os.path.join(config.dir_accounts_fp, uid+'.json')
    if interactive:
        pr = f'\nDo you want to delete or wipe out the "{user}" account \
(<enter> to cancel) [d/w]? '
        go = input(pr)
    else:
        go = action
    if go.lower() == 'd':
        bak_file = '{0}.{1}-{2}'.format(account_file_fp, 'removed', \
                                        int(time.time()))
        os.rename(account_file_fp, bak_file)
        if not args.quiet:
            print(f'{nl}The "{user}" account has been deleted and renamed:')
            print(bak_file)
        if not interactive: sys.exit()
    elif go.lower() == 'w':
        os.remove(account_file_fp)
        if not args.quiet:
            print(f'{nl}The "{user}" account has been wiped out.')
        if not interactive: sys.exit()
    else:
        print('\nOK, nothing was done.')
    return

#############################################################################

def list_users(interactive=True):
    """
list_users(interactive[bool])

    List koi users.

"""
    if not profiles:
        if interactive:
            print('\nNo users found.')
        else:
            print('No users found.')
        return
    if interactive:
        print(f'\nAccounts directory: {config.dir_accounts_fp}')
    for uid, pf in profiles.items():
        if pf['multiF_check']['enabled']:
            status = 'enabled'
        else:
            status = 'disabled'
        print(f'{pf["user"]}: name={pf["name"]}, locked={pf["locked"]}, \
multi-factor={status}, roles={pf["roles"]}, uid={uid}')
    if not interactive:
        sys.exit()
    while True:
        user = input('\nEnter a username for details (<enter> to quit): ')
        if not user:
            break
        uid = hashlib.md5(user.lower().encode()).hexdigest()
        if profiles.get(uid, ''):
            print('')
            pprint.pprint(profiles[uid])
        else:
            print(f'User "{user}" does not exist.')
    return

#############################################################################

def mod_account(interactive=True, args=None):
    """
mod_account(interactive[bool], args[namespace])

    Modify an account.

"""
    if interactive:
        print('\nModifying account')
        user = input('\nUser name (<enter> to cancel): ')
    else:
        user = args.user
    if not user:
        return
    uid = hashlib.md5(user.lower().encode()).hexdigest()
    account_file_fp = os.path.join(config.dir_accounts_fp, uid+'.json')
    if uid not in profiles:
        if interactive:
            print(f'\nAccount "{user}" does not exist.')
            return
        else:
            print(f'Account "{user}" does not exist.')
            sys.exit(1)
    save = True
    if not interactive:
        if args.firstname: profiles[uid]['full_name']['first'] = args.firstname
        if args.middlename: profiles[uid]['full_name']['middle'] = args.middlename
        if args.lastname: profiles[uid]['full_name']['last'] = args.lastname
        if args.name:
            profiles[uid]['name'] = args.name
        else:
            profiles[uid]['name'] = ' '.join(f'{profiles[uid]["full_name"]["first"]} {profiles[uid]["full_name"]["middle"]} {profiles[uid]["full_name"]["last"]}'.split())
        if args.primary_address: profiles[uid]['address']['primary'] = args.primary_address
        if args.primary_email: profiles[uid]['email']['primary'] = args.primary_email
        if args.primary_phone: profiles[uid]['phone']['primary'] = args.primary_phone
        if args.password:
            if not args.password.startswith('$6$') and \
               not args.password.startswith('$y$'):
                print('Only SHA512 and yescrypt hashes are allowed.')
                sys.exit(1)
            profiles[uid]['hash'] = args.password
        pg = []
        if args.group == []:
            profiles[uid]['groups'] = []
        elif args.group:
            pg = [i.strip().lower() for i in args.group]
        if args.agroup:
            pg = profiles[uid]['groups'] + \
                [i.strip().lower() for i in args.agroup]
        if pg:
            profiles[uid]['groups'] = sorted(list(set(pg)))
        if args.lock: profiles[uid]['locked'] = True
        if args.unlock: profiles[uid]['locked'] = False
        if args.twof_on: profiles[uid]['multiF_check']['enabled'] = True
        if args.twof_off: profiles[uid]['multiF_check']['enabled'] = False
        if args.twof_email:
            profiles[uid]['multiF_check']['twoF_email'] = args.twof_email
        if args.editor_off:
            if 'editor' in profiles[uid]['roles'][config.tpl_edit]:
                profiles[uid]['roles'][config.tpl_edit].remove('editor')
        if args.editor_on:
            if 'editor' not in profiles[uid]['roles'][config.tpl_edit]:
                profiles[uid]['roles'][config.tpl_edit].append('editor')
    else:
        print(f'\nName: {profiles[uid]["name"]}, details below.')
        under = "="*len(profiles[uid]["name"])
        print(f'           {under}')
        pprint.pprint(profiles[uid])
        print('')
        while True:
            print('0) Set name/address/email/phone')
            print('1) Add/set groups')
            print('2) Toggle edit role')
            print('3) Toggle account lock')
            print('4) Set password')
            print('5) Set two-factor email')
            print('6) Set/render two-factor TOTP (time-based one-time password)')
            print('7) Toggle two-factor')
            print('8) Return')
            field = input('\nSelection: ')
            if field in [f'{i}' for i in range(0,9)]:
                break
            else:
                print("\nSorry, I don't understand.\n")
        if field == '0':
            print("\nIn the fields below press <enter> to skip, '?' to blank out.")
            first_name = input(f'\nUser first name [{profiles[uid]["full_name"]["first"]}]: ')
            if first_name == '?':
                profiles[uid]['full_name']['first'] = ''
            elif first_name:
                profiles[uid]['full_name']['first'] = first_name
            print(f'First name will be: {profiles[uid]["full_name"]["first"]}')
            middle_name = input(f'User middle name(s) [{profiles[uid]["full_name"]["middle"]}]: ')
            if middle_name == '?':
                profiles[uid]['full_name']['middle'] = ''
            elif middle_name:
                profiles[uid]['full_name']['middle'] = middle_name
            print(f'Middle name will be: {profiles[uid]["full_name"]["middle"]}')
            last_name = input(f'User last name [{profiles[uid]["full_name"]["last"]}]: ')
            if last_name == '?':
                profiles[uid]['full_name']['last'] = ''
            elif last_name:
                profiles[uid]['full_name']['last'] = last_name
            print(f'Last name will be: {profiles[uid]["full_name"]["last"]}')
            profiles[uid]['name'] = ' '.join(f'{profiles[uid]["full_name"]["first"]} {profiles[uid]["full_name"]["middle"]} {profiles[uid]["full_name"]["last"]}'.split())
            print(f"\nName is (for user {user}): {profiles[uid]['name']}")
            address = input(f'\nUser primary address [{profiles[uid]["address"]["primary"]}]: ')
            if address == '?':
                profiles[uid]['address']['primary'] = ''
            elif address:
                profiles[uid]['address']['primary'] = address
            print(f'Primary address will be: {profiles[uid]["address"]["primary"]}')
            email = input(f'User primary email [{profiles[uid]["email"]["primary"]}]: ')
            if email == '?':
                profiles[uid]['email']['primary'] = ''
            elif email:
                profiles[uid]['email']['primary'] = email
            print(f'Primary email will be: {profiles[uid]["email"]["primary"]}')
            phone = input(f'User primary phone [{profiles[uid]["phone"]["primary"]}]: ')
            if phone == '?':
                profiles[uid]['phone']['primary'] = ''
            elif phone:
                profiles[uid]['phone']['primary'] = phone
            print(f'Primary phone will be: {profiles[uid]["phone"]["primary"]}')
        elif field == '1':
            print('\nNote that groups are lowercased, so "AV" becomes "av".')
            print('You will also be given the option to use or add the groups \
you list.')
            print(f'Current groups: {" ".join(profiles[uid]["groups"])}')
            groups = input('Space-separated list of groups (<enter> for none): ')
            if groups:
                add = input('Add to existing list (will otherwise replace) [Y/n]? ')
            else:
                add = 'n'
            if add.lower() == 'y':
                pg = profiles[uid]['groups'] + \
                    [i.strip().lower() for i in groups.split()]
            else:
                pg = [i.strip().lower() for i in groups.split()]
            profiles[uid]['groups'] = sorted(list(set(pg)))
            print(f"\nGroups are now ({user}): {profiles[uid]['groups']}")
        elif field == '2':
            if 'editor' in profiles[uid]['roles'].get(config.tpl_edit, []):
                is_editor = True
                action = input('\nDisable editing role [y/N]?: ')
            else:
                is_editor = False
                action = input('\nEnable editing role [y/N]?: ')
            if is_editor and action.lower() == 'y':
                profiles[uid]['roles'][config.tpl_edit].remove('editor')
                print(f'\nUser "{user}" is no longer an editor.')
            elif not is_editor and action.lower() == 'y':
                profiles[uid]['roles'][config.tpl_edit].append('editor')
                print(f'\nUser "{user}" is now an editor.')
            else:
                save = False
        elif field == '3':
            if profiles[uid]['locked']:
                action = input('\nUnlock account [y/N]?: ')
            else:
                action = input('\nLock account [y/N]?: ')
            if profiles[uid]['locked'] and action.lower() == 'y':
                profiles[uid]['locked'] = False
                print(f'\nAccount "{user}" is unlocked.')
            elif not profiles[uid]['locked'] and action.lower() == 'y':
                profiles[uid]['locked'] = True
                kill = input('Terminate current session (if any) [y/N]? ')
                if kill.lower() == 'y':
                    profiles[uid]['token'] = ''
                    profiles[uid]['nonce'] = ''
                    profiles[uid]['xCSRF'] = ''
                print(f'\nAccount "{user}" is locked.')
            else:
                save = False
        elif field == '4':
            pwhash = get_password(user)
            profiles[uid]['hash'] = pwhash
            print(f'\nPassword has been updated ({user}).')
        elif field == '5':
            print('\nLeave blank to disable 2F emails.')
            email = get_email()
            profiles[uid]['multiF_check']['enabled'] = True
            profiles[uid]['multiF_check']['twoF_email'] = email
            if email:
                print(f'\nTwo-factor email has been set and enabled ({user}).')
            else:
                print(f'\nTwo-factor email has been disabled ({user}).')
        elif field == '6':
            if (twoF_totp := profiles[uid]['multiF_check']['twoF_totp']):
                pr = '\nTOTP and QR code: render, delete, skip? [r/d/S]? '
                totp = input(pr)
            else:
                pr = '\nEnable TOTP and render QR code ("s" to skip)? [y/S]? '
                totp = input(pr)
            if totp.lower() == 'r':
                totp = TOTP.using(secrets=config.twoF_totp_secret).from_source(twoF_totp)
                print(pyqrcode.create(totp.to_uri()).terminal(quiet_zone=1))
                print(f'Alternative non-QR key ({user}): {totp.pretty_key()}')
                save = False
            elif totp.lower() == 'd':
                profiles[uid]['multiF_check']['twoF_totp'] = ''
                print(f'\nTwo-factor TOTP has been removed ({user}).')
            elif totp.lower() == 'y':
                totp = TOTP.using(secrets=config.twoF_totp_secret).new(issuer=config.twoF_totp_issuer, \
                                                                       label=profiles[uid]['user'])
                profiles[uid]['multiF_check']['enabled'] = True
                profiles[uid]['multiF_check']['twoF_totp'] = totp.to_json()
                print('\nPlease scan the following QR code with your OTP app:')
                print(pyqrcode.create(totp.to_uri()).terminal(quiet_zone=1))
                print(f'Alternatively, you can use the following key: {totp.pretty_key()}')
                print(f'\nTwo-factor TOTP has been set and enabled ({user}).')
            else:
                save = False
        elif field == '7':
            if profiles[uid]['multiF_check']['enabled']:
                action = input('\nDisable multi-factor [y/N]?: ')
            else:
                action = input('\nEnable multi-factor [y/N]?: ')
            if profiles[uid]['multiF_check']['enabled'] and action.lower() == 'y':
                profiles[uid]['multiF_check']['enabled'] = False
                profiles[uid]['nonce'] = ''
                print(f'\nMulti-factor login disabled ({user}).')
            elif not profiles[uid]['multiF_check']['enabled'] and action.lower() == 'y':
                profiles[uid]['multiF_check']['enabled'] = True
                print(f'\nMulti-factor login enabled ({user}).')
            else:
                save = False
        else:
            return
        if save:
            if not profiles[uid]['multiF_check']['twoF_email'] and not \
               profiles[uid]['multiF_check']['twoF_totp']:
                profiles[uid]['multiF_check']['enabled'] = False
            print(f'\nSaving {account_file_fp} ({user})')
    if save:
        with open(account_file_fp, 'w', encoding='utf-8') as fd:
            json.dump(profiles[uid], fd, ensure_ascii=False)
        os.chmod(account_file_fp, config.wwwfperms)
        if getpass.getuser() == 'root':
            shutil.chown(account_file_fp, config.wwwuid, config.wwwgid)
    if not interactive and not args.quiet:
        print(f'Account "{user}" has been modified.')
        sys.exit()
    return

#############################################################################

def batch_import():
    """
batch_import()

    Import accounts from passwd and shadow files.

"""
    print('\nBatch import')
    def_passwd_file = '/etc/passwd'
    def_shadow_file = '/etc/shadow'
    passwd_file = input(f'\n"passwd" file (full path) [{def_passwd_file}]: ')
    if not passwd_file:
        passwd_file = def_passwd_file
    print('Note: only SHA512 and yescrypt password hashes are supported, \
account will be locked otherwise.')
    shadow_file = input(f'"shadow" file (full path) [{def_shadow_file}]: ')
    if not shadow_file:
        shadow_file = def_shadow_file
    if shadow_file == def_shadow_file and getpass.getuser() != 'root':
        print(f'\nYou need to be root to access "{shadow_file}"')
        return
    uid_lim = input('Starting UID to import: ')
    print('If you have a "user email" two-column list please enter it below.')
    emailF_fp = input('Email list path (<enter> for none): ')
    if emailF_fp:
        twoF_use = input('Use the emails for two-factor [y/N]? ')
    else:
        twoF_use = 'n'
    emailF_users = {}
    if emailF_fp:
        with open(emailF_fp) as f:
            for idx, entry in enumerate(f):
                (user, email) = entry.strip().split()
                emailF_users[user] = email
    shadow_dict = {}
    with open(shadow_file) as f:
        for idx, entry in enumerate(f):
            fields = entry.split(':')
            shadow_dict[fields[0]] = fields[1]
    count = 0
    with open(passwd_file) as f:
        for idx, entry in enumerate(f):
            fields = entry.split(':')
            uid = int(fields[2])
            if uid < int(uid_lim):
                continue
            user = fields[0]
            kuid = hashlib.md5(user.lower().encode()).hexdigest()
            if kuid in profiles:
                print(f'"{user}" account already exists, skipping.')
                continue
            email = {'primary': ''}
            twoF_email = ''
            if user in emailF_users:
                email['primary'] = emailF_users[user]
                if twoF_use.lower() == 'y':
                    twoF_email = email['primary']
            gecos = fields[4]
            name = gecos.split(',')[0]
            pwhash = shadow_dict[user]
            if not pwhash.lstrip('!').startswith('$6$') and \
               not pwhash.lstrip('!').startswith('$y$'):
                print(f'"{user}" account does not use either SHA-512 or yescrypt, locking.')
                pwhash = '!'
                locked = True
            else:
                locked = False
            if twoF_email:
                multiF_check = {'enabled': True, 'twoF_email': twoF_email, \
                                'twoF_totp': ''}
            else:
                multiF_check = {'enabled': False, 'twoF_email': '', \
                                'twoF_totp': ''}
            profiles[kuid] = {'user': user, 'name': name, 'uid': kuid, \
                              'full_name': {'first': '', 'middle': '', 'last': ''}, \
                              'gecos': gecos, 'token': '', 'groups': [], 'nonce': '', \
                              'xCSRF': '', 'data': {}, 'address': {'primary': ''}, \
                              'email': email, 'phone': {'primary': ''}, \
                              'locked': locked, 'roles': {config.tpl_edit: []}, \
                              'cache': {}, 'login_ts': 0, 'logout_ts': 0, 'ip': None, \
                              'knock_tss': [0], 'koi_version': __version__, \
                              'created': int(time.time()), \
                              'hash': pwhash, 'multiF_check': multiF_check}
            account_file_fp = os.path.join(config.dir_accounts_fp, kuid+'.json')
            print(f'Saving "{user}" in: {account_file_fp}')
            with open(account_file_fp, 'w', encoding='utf-8') as fd:
                json.dump(profiles[kuid], fd, ensure_ascii=False)
            os.chmod(account_file_fp, config.wwwfperms)
            if getpass.getuser() == 'root':
                shutil.chown(account_file_fp, config.wwwuid, config.wwwgid)
            count+=1
        print(f'Accounts created: {count}')
    return

#############################################################################

def exit():
    sys.exit()

#############################################################################

if __name__ == '__main__':

    import config
    import glob
    import os
    import json
    import hashlib
    import pwd
    import spwd
    import pwd
    import getpass
    import platform
    import pprint
    import time
    import secrets
    import shutil
    import argparse
    import sys
    import pyqrcode
    from passlib.hash import sha512_crypt
    from passlib.totp import TOTP
    import logging
    logging.disable(logging.INFO)
    from koi import __version__

    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--add', \
                        help='add new user', \
                        action='store_true')
    parser.add_argument('-d', '--delete', \
                        help='delete user (will back-up account)', \
                        action='store_true')
    parser.add_argument('-e', '--editor_on', \
                        help='enable editor role', \
                        action='store_true')
    parser.add_argument('-xe', '--editor_off', \
                        help='disable editor role', \
                        action='store_true')
    parser.add_argument('-fn', '--firstname', \
                        help="user's first name")
    parser.add_argument('-g', '--group', \
                        help='set groups (space-separated list, none to empty)', \
                        action="extend", nargs="*", type=str)
    parser.add_argument('-G', '--agroup', \
                        help='add groups (space-separated list)', \
                        action="extend", nargs="*", type=str)
    parser.add_argument('-ln', '--lastname', \
                        help="user's last name")
    parser.add_argument('-L', '--lock', \
                        help='lock account', \
                        action='store_true')
    parser.add_argument('-m', '--twof_email', \
                        help='set two-factor email address')
    parser.add_argument('-mn', '--middlename', \
                        help="user's middle name(s)")
    parser.add_argument('-n', '--name', \
                        help='set name (construct using "fn"/"mn"/"ln" otherwise)')
    parser.add_argument('-o', '--overview', \
                        help='overview of all accounts', \
                        action='store_true')
    parser.add_argument('-p', '--password', \
                        help='set password (using hash)')
    parser.add_argument('-pa', '--primary_address', \
                        help="set user's primary address")
    parser.add_argument('-pe', '--primary_email', \
                        help="set user's primary email")
    parser.add_argument('-pp', '--primary_phone', \
                        help="set user's primary phone")
    parser.add_argument('-q', '--quiet', \
                        help='quiet mode', \
                        action='store_true')
    parser.add_argument('-s', '--show',
                        help='show profile keys (list, none for all)',
                        nargs='*')
    parser.add_argument('-t', '--twof_on', \
                        help='enable multi-factor', \
                        action='store_true')
    parser.add_argument('-xt', '--twof_off', \
                        help='disable multi-factor', \
                        action='store_true')
    parser.add_argument('-u', '--user', \
                        help='user name')
    parser.add_argument('-U', '--unlock', \
                        help='unlock account', \
                        action='store_true')
    parser.add_argument('-w', '--wipe', \
                        help='wipe out user account', \
                        action='store_true')

    args = parser.parse_args()
    profiles = get_profiles()
    if any(vars(args).values()):
        if (args.add and args.delete) or (args.add and args.wipe) or \
                        (args.delete and args.wipe):
            print('You can only choose one of "add", "delete", and "wipe".')
            sys.exit(1)
        elif args.lock and args.unlock:
            print('You can only choose one of "lock" and "unlock".')
            sys.exit(1)
        elif args.group and args.agroup:
            print('You can only choose one of "set groups" and "add groups".')
            sys.exit(1)
        elif args.twof_on and args.twof_off:
            print('You can only choose one of "twof_on" and "twof_off".')
            sys.exit(1)
        elif args.editor_on and args.editor_off:
            print('You can only choose one of "editor_on" and "editor_off".')
            sys.exit(1)
        elif args.overview:
            list_users(interactive=False)
        elif not args.user:
            print('You must specify a user to act upon.')
            sys.exit(1)
        elif args.show or args.show == []:
            uid = hashlib.md5(args.user.lower().encode()).hexdigest()
            if uid not in profiles:
                print(f'Account "{args.user}" does not exist.')
                sys.exit(1)
            if args.show == []:
                pprint.pprint(profiles[uid])
                sys.exit()
            for i in args.show:
                if i in profiles[uid]:
                    print(profiles[uid][i])
                else:
                    print(f'"{i}" is not a profile key.')
        elif args.delete:
            rem_account(interactive=False, user=args.user, action='d')
        elif args.wipe:
            rem_account(interactive=False, user=args.user, action='w')
        elif args.add:
            add_account(interactive=False, args=args)
        else:
            mod_account(interactive=False, args=args)
    else:
        options = {'0': ('Search',             'search_accounts()' ),
                   '1': ('Add account',        'add_account()'     ),
                   '2': ('Remove account',     'rem_account()'     ),
                   '3': ('Modify account',     'mod_account()'     ),
                   '4': ('List users/account', 'list_users()'      ),
                   '5': ('Batch import',       'batch_import()'    ),
                   '6': ('Exit',               'exit()'            )}

        print('\nkoi account manager (use "-h" for non-interactive usage)')
        print('========================================================')
        while True:
            print('')
            for key, value in options.items():
                print(f'{key}) {value[0]}')
            n = input('\nSelection: ')
            try:
                eval(options[n][1])
                # Refresh the database...
                profiles = get_profiles()
            except Exception as e:
                print(f'\nThere was an error, something about: {e}')
            except SystemExit:
                print('\nGoodbye')
                break
